#include "mex.h"
#include <math.h>

/* global constants */

/* lambda value 
const double lambdaValue = 5.41; */
/* sequence for lambda values for 3D images*/
const double lambdaSeq3D[15] = {1, 1.9, 1.5, 1.3, 1.3, 1.3, 1.3, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1, 1.1}; 

/* Calculates the location kernel value of a given value */
double locationKernel(double x){
    if (x > 1 || x < 0) 
        return 0;    
	else 
        return (1-x);	 
}

/* Calculates the statistical kernel value of a given value */
double statisticalKernel(double x){
    if (x > 1 || x < 0)
        return 0;
    if (x < 1/4.) 
        return 1;
    else
        return 4/3. * (1-x);
}

int getStepsNumber(double hInit, double hIncr, double hMax) {
    int n = 0;
    double hCurr = hInit;
    while (hCurr<hMax) {
        hCurr = hCurr*hIncr;
        n++;
    }
    return n;
}

double getHCurrAtStepNr(int i, double hInit, double hIncr) {
     int j;
     double hCurr = hInit;
     for (j=0; j<i; j++) {
         hCurr = hCurr*hIncr;
     }
     return hCurr;
}

/*bool isnan(double value)
{
    return ((*(unsigned int*)&value) & 0x7fffffff) > 0x7f800000;
}*/


/* This function smooths a three-dimensional array using variance estimates */
void smoothAdaptive(double* thetaOriginal, double* variancesOriginal, double* residualsOriginal, const int* dim, double hMax, double lambda, double* voxelSizes, double* thetaNew, double* variancesNew, double* residualsNew, int nscan){
	
    int i, ix, iy, iz, jx, jy, jz;
    int minX, minY, minZ, maxX, maxY, maxZ;
    int ihX, ihY, ihZ, dlwX, dlwY, dlwZ, clwX, clwY, clwZ;
    
    int iscan;
    
    	/* EG */
    double* thetaOld;
    double* sumOfWeights;
    double lambdaCurrent;
    int iind;
    double ni;
    double thni;
    double vari;
    double* resiTime;
	
    
    int k = 0;
    double hInit = 1;
	double hCurr = hInit;
 	double hCurrSq = hCurr*hCurr;
	double hIncr = pow(1.25, 0.3333333);
    int    nIteration = getStepsNumber(hInit,hIncr,hMax);
     
    /* number of voxel */
	int numberOfVoxel = dim[0]*dim[1]*dim[2];

    /* squared voxelsizes*/
    double voxelSizeSq[3];
    voxelSizeSq[0] = voxelSizes[0]*voxelSizes[0];	
	voxelSizeSq[1] = voxelSizes[1]*voxelSizes[1];	
	voxelSizeSq[2] = voxelSizes[2]*voxelSizes[2];	

    /* estimates for theta at next step */
    for (i = 0; i < numberOfVoxel; i++) {
        thetaNew[i] = thetaOriginal[i]; 
        variancesNew[i] = variancesOriginal[i];  
    }
    
    /* estimates for theta at current step, 
     * initialize with original data from thetaNew, 
     * when entering the iteration */
    thetaOld = (double*) malloc(numberOfVoxel*sizeof(double));
    
    /* sum of weights */    
	sumOfWeights = (double*) malloc(numberOfVoxel*sizeof(double));
	for (i = 0; i < numberOfVoxel; i++) sumOfWeights[i] = 1;
    
   
    resiTime = (double*) malloc(nscan*sizeof(double));
    
       
	/* main iteration loop */
    
    for (k = 0; k< nIteration; k++) {
	    
        hCurr = getHCurrAtStepNr(k, hInit, hIncr);
        hCurrSq = hCurr*hCurr;
        /*mexPrintf("using hCurr %f ... \n", hCurr);
        mexPrintf("using hCurrSq %f ... \n", hCurrSq);*/
        
        /* new estimates are now the old ones */
		for (i = 0; i < numberOfVoxel; i++) { 
			thetaOld[i] = thetaNew[i];
		}	

        /* calculate adaptation bandwidth lambda for current step 
         * from the global(!) constant lambdaValue with a sequence 
         * of corrections */
        double lambdaCurrent = lambda;
		if (k <= 15)
            lambdaCurrent *= lambdaSeq3D[k];

        /*printf("hCurr %f\n", hCurr);
        printf("voxelSizes %f %f %f\n", voxelSizes[0], voxelSizes[1], voxelSizes[2]);*/
        
        ihX = ceil(hCurr / voxelSizes[0]);
        ihY = ceil(hCurr / voxelSizes[1]);
        ihZ = ceil(hCurr / voxelSizes[2]);
        
        /* for each voxel i ... */
        for (iz = 0; iz < dim[2]; iz++) {
			for (iy = 0; iy < dim[1]; iy++) {
                for (ix = 0; ix < dim[0]; ix++) {

                    iind = ix + iy*dim[0] + iz*dim[0]*dim[1]; /* vector index of voxel i */
                    ni = 0; /* sum of weights in voxel i */
                    thni = 0; /* estimated theta in voxel i */
                    vari = 0; /* estimated variance in voxel i */
                    

                    for (iscan = 0; iscan < nscan; iscan++) {
                        resiTime[iscan] = 0;                                        
                    }
                    
                    minX = ix - ihX;
                    minY = iy - ihY;
                    minZ = iz - ihZ;
                    maxX = ix + ihX;
                    maxY = iy + ihY;
                    maxZ = iz + ihZ;

                    if (minX < 0) minX = 0;
                    if (minY < 0) minY = 0;
                    if (minZ < 0) minZ = 0;
                    if (maxX >= dim[0]) maxX = dim[0] - 1;
                    if (maxY >= dim[1]) maxY = dim[1] - 1;
                    if (maxZ >= dim[2]) maxZ = dim[2] - 1;
                   
                 
                    
                    /* for each voxel j ... */
                    for (jz = minZ; jz <= maxZ; jz++) {
                        for (jy = minY; jy <= maxY; jy++) {
                            for (jx = minX; jx <= maxX; jx++) {
                                
                                int jind = jx + jy*dim[0] + jz*dim[0]*dim[1]; /* vector index of voxel j */
                                double sqD = voxelSizeSq[0]*(ix-jx)*(ix-jx) + voxelSizeSq[1]*(iy-jy)*(iy-jy) + voxelSizeSq[2]*(iz-jz)*(iz-jz);
                                if ((sqD <= hCurrSq) & (!isnan(thetaOriginal[jind]))){
                                    double thetaDiff = thetaOld[iind] - thetaOld[jind];
                                    double sij = sumOfWeights[iind] * thetaDiff * thetaDiff / (2 * variancesOriginal[iind]);
                                    double wij = locationKernel(sqD/hCurrSq) * statisticalKernel(sij/lambdaCurrent);
                                    ni += wij;
                                    thni += wij * thetaOriginal[jind];
                                    /*resi += wij * residualsOriginal[jind];*/
                                    vari += wij * wij * variancesOriginal[iind];
                                    if (k==nIteration-1){
                                        for (iscan = 0; iscan < nscan; iscan++) {
                                        resiTime[iscan] += wij * residualsOriginal[iscan+jind*nscan];                                         
                                        }
                                    }
                                }
                                                                                 
                               
                            } /* for jx */
                        } /* for jy */
                    } /* for jz */

                    thetaNew[iind] = thni/ni; /* ni should never be zero, because wij should be 1 for i==j */
                    /*residualsNew[iind] = resi/ni;*/
                    sumOfWeights[iind] = ni; /* we no longer need the sum of weights from previous step in voxel i */
                    variancesNew[iind] = vari/ni/ni;
                    if (k==nIteration-1){
                        for (iscan = 0; iscan < nscan; iscan++) {                            
                                residualsNew[iscan+iind*nscan] = resiTime[iscan]/ni;
                        }
                    }
                    

                } /* for iz */
			} /* for iy */
		} /* for ix */
			
   

		mexPrintf("finished step %d ... \n", k);
        mexEvalString("pause(.001);");
        
        
    } /* main iteration loop */
    
    /* EG */
	if(thetaOld!=NULL) 
		free(thetaOld);
	
	if(sumOfWeights!=NULL)
		free(sumOfWeights);
	
	if(resiTime!=NULL)
		free(resiTime);
	

}



/* 
* Entry function.
* Called from spm_smooth_gammas.m
*/
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]){

	const int* dim;
    const int* dim2;
	double* voxelSizes;
	double voxelMin, hMax, lambda;
	double *thetaOriginal, *variancesOriginal, *residualsOriginal, *thetaNew, *variancesNew, *residualsNew;
    double *test;
    int nscan;

    if (nlhs > 3) mexErrMsgTxt ("Too many output arg."); /* to be adjusted */
    if (nrhs != 7) mexErrMsgTxt ("Wrong number of input arg.");  /* to be adjusted */

    /* get the input arguments */
    dim               = mxGetDimensions(prhs[0]);    
	thetaOriginal     = (double*) mxGetPr(prhs[0]);	
	variancesOriginal = (double*) mxGetPr(prhs[1]);
    residualsOriginal = (double*) mxGetPr(prhs[2]);
	hMax              = mxGetScalar(prhs[3]);
	lambda            = mxGetScalar(prhs[4]);
    voxelSizes        = (double*) mxGetPr(prhs[5]);
    nscan             = (int)mxGetScalar(prhs[6]);
    dim2               = mxGetDimensions(prhs[2]); 

    /* the minimal voxel edge will be set to one 
     * then hMax is measured in units of the smallest edge */
	if (voxelSizes[0] < voxelSizes[1]) 
        voxelMin = voxelSizes[0];
	else 
        voxelMin = voxelSizes[1];
	if (voxelSizes[2] < voxelMin) 
        voxelMin = voxelSizes[2];
	voxelSizes[0] /= voxelMin;
	voxelSizes[1] /= voxelMin;
	voxelSizes[2] /= voxelMin;

	/* output arrays */
    plhs[0]     = mxCreateNumericArray (3, dim, mxDOUBLE_CLASS, mxREAL); 
    thetaNew = (double*) mxGetPr (plhs[0]);
    plhs[1]     = mxCreateNumericArray (3, dim, mxDOUBLE_CLASS, mxREAL); 
    variancesNew = (double*) mxGetPr (plhs[1]);
    plhs[2]     = mxCreateNumericArray (4, dim2, mxDOUBLE_CLASS, mxREAL); 
    residualsNew = (double*) mxGetPr (plhs[2]);
	
	/* main call to adaptive smoothing procedure */
	smoothAdaptive(thetaOriginal, variancesOriginal, residualsOriginal, dim, hMax, lambda, voxelSizes, thetaNew, variancesNew, residualsNew, nscan);

}
