function [hreg,xSPM,SPM]=spm_smooth_gamma(spm_filename, hMax, lambda)
% adaptative weighted smoothing for the betas
%_______________________________________________________________________
% Copyright (C) 2016 WIAS Berlin
%
% To get information about this toolbox:
% http://www.wias-berlin.de/software/imaging/aws4SPM.jsp
%
% 

SPM=[];
load(spm_filename{:});

% save the current folder location
originalDir = pwd;

%-Check if the model has been estimated
%--------------------------------------------------------------------------
try
    SPM.xVol.S;
catch
        
    str = { 'This model has not been estimated.';...
        'Would you like to estimate it now?'};
    if spm_input(str,1,'bd','yes|no',[1,0],1)
        SPM.swd      = spm_file(spm_filename{:},'fpath');
        SPM = spm_spm(SPM);
    else
        SPM = []; xSPM = [];  hreg=[];
        
        return
    end
end




% save original SPM
originalSPM = SPM;

try
    
    % in case the toolbox has already been used
    % take the contrasts that were defined in aws4SPM 
    % if not initialize the additional variables
    if isfield(SPM,'aws_xCon'),
        SPM.xCon = SPM.aws_xCon;  
        VRes = SPM.aws_VRes;
    else
        SPM.awsGamma = struct('dim',   [],...
            'dt',     [],...
            'mat',      [],...
            'pinfo',   [],...
            'fname',     [],...
            'descrip',       [],...
            'n', [],...
            'private', []);
        SPM.awsRes = struct('dim',   [],...
            'mat',      [],...
            'pinfo',   [],...
            'fname',     [],...
            'descrip',       [],...
            'n', [],...
            'dt',     [],...
            'private', []);
         SPM.xCon = {};
        SPM.aws_VRpv = struct('fname',     [],...
            'dim',   [],...
            'dt',     [],...
            'mat',      [],...
            'pinfo',   [],...
            'descrip',       [],...
            'n', [],...            
            'private', []);       
        SPM.aws_FWHM = {};
        SPM.aws_R = {};
        % write residuals file if not present or create an handle to them        
        if ~spm_existfile('Res_0001.nii')
            VRes = spm_write_residuals(SPM,NaN);
        else
            VRes = struct('fname', []);
            for j=1:SPM.nscan,   
                     VRes(j).fname = sprintf('Res_%04d.nii', j);   
            end
        end            
        
    end
    old_NumberAwsContrasts = length(SPM.xCon);
    
    % read the absolute path saved in SPM.mat
    % this must contain all the necessary files
    % (in particular beta_???.<ext> and ResMS.<ext>)
    swd = SPM.swd;
    
    % we have to change dir before calling spm_conman or a new SPM.mat will
    % be saved in the current directory
    cd(swd);
    
    
    % define and/or select new contrast 
    % possible to define more than one new contrasts, only one can be
    % selected to run the computation
    [I, xCon] = spm_conman(SPM, 'T', 1, ...
        '    Select contrasts...', '', 1); % 'for conjunction' taken away
    % Now I contains the selected contrast in the struct xCon
    
    SPM.xCon = xCon; %
    new_NumberAwsContrasts = length(SPM.xCon);
    
    
    
    % if the contrast selected has already been used in a computation in
    % aws, gives the choice to just show the old computation or overwrite
    % files and related fields
    if ~isempty(SPM.xCon(I).Vcon)
        if ~isempty(strfind(SPM.xCon(I).name,'lambda')),
            str = {'This contrast has already been used to perform the smoothing.',...
                'If you stop, existing results will be displayed.',...
                'If you continue, existing results will be overwritten!',...
                'Please restart aws to define a new contrast with the same value, but using different lambda and hMax.'};
            if spm_input(str,1,'bd','stop|continue',[1,0],1)
                spm('Pointer','Arrow')
                
                % save in the variables that are used by results
                SPM.xVol.FWHM =  originalSPM.aws_FWHM{I};        
                SPM.xVol.R =  originalSPM.aws_R{I};
                SPM.xVol.VRpv =  originalSPM.aws_VRpv(I);
                
                [hreg,xSPM,SPM] = spm_results_aws('Setup',SPM,I);
                 cd(originalDir);
                return
            else
                sw = warning('off','backtrace');
                warning('Overwriting old results\n\t (pwd = %s) ',SPM.swd);
                warning(sw);
                % empty the fields
                SPM.xCon(I).eidf =[];
                SPM.xCon(I).Vcon =[];
                SPM.xCon(I).Vspm =[];
                
                % clear the information about the old lambda and hmax in
                % the name
                SPM.xCon(I).name = regexprep(SPM.xCon(I).name,'[=][\s]\d*[.]\d*','');
                SPM.xCon(I).name = regexprep(SPM.xCon(I).name,'[=][\s]\d*','');
                SPM.xCon(I).name = strrep(SPM.xCon(I).name,' (lambda , hMax )','');
                
                
            end
            
            
        end
    end
    
    if isempty(SPM.xCon(I).Vcon),
        
        % add information about aws, lambda and hMax in the name of the
        % contrast
        
        for i=(old_NumberAwsContrasts+1):new_NumberAwsContrasts
            SPM.xCon(i).name = strcat(SPM.xCon(i).name,' - aws');
        end
        
        contrastName = sprintf(' (lambda = %1.2f, hMax = %1.2f)',lambda,hMax);
        SPM.xCon(I).name = strcat(SPM.xCon(I).name,contrastName);
        
        % find the contrast data
        
        contrast = SPM.xCon(I).c;
        
        nrContr = size(contrast);
        
        % voxel dimensions in mm
        voxelSizes = abs(diag(SPM.xY.VY(1).mat(1:3,1:3)))';
        
        % this should be (X'X)^{-1}
        covM = SPM.xX.Bcov;
        
        % degrees of freedom
        dof = SPM.xX.erdf;
        
        % dimension of datacube (3D)
        DIM = SPM.xY.VY(1).dim(1:3)';
        xDim = DIM(1,1);
        yDim = DIM(2,1);
        zDim = DIM(3,1);
        numberOfVoxel = xDim*yDim*zDim;
        
        % create position matrix
        % this matrix determines the in-mask brain voxels in this case: all
        % we could also try to use SPM.xVol.XYZ which seems to contain a
        % brain mask automatically
        XYZ(1, numberOfVoxel) = xDim;
        XYZ(2, numberOfVoxel) = yDim;
        XYZ(3, numberOfVoxel) = zDim;
        counter = 1;
        for l = 1:zDim
            for j = 1:yDim
                for i = 1:xDim
                    XYZ(1, counter) = i;
                    XYZ(2, counter) = j;
                    XYZ(3, counter) = l;
                    counter = counter+1;
                end
            end
        end
        
        % read in betas from fmri GLM
        % calculate contrast image gamma = c' * beta
        gammaData = zeros(1, numberOfVoxel);
        for z = 1:nrContr(1)
            if contrast(z, 1) ~= 0
                fileName = SPM.Vbeta(z).fname;
                betaCurr = spm_get_data(fileName, XYZ);
                gammaData = contrast(z, 1) * betaCurr + gammaData;
            end
        end
        gamma = reshape(gammaData, xDim, yDim, zDim);
        
        % determine the variances
        % Var = xBold*ResMS/dof
        % read-in Residual Mean Squares
        fileName = SPM.VResMS.fname;
        ResMSData = spm_get_data(fileName, XYZ);
        ResMS = reshape(ResMSData, xDim, yDim, zDim);
        
        % divide ResMS by degrees of freedom
        % exchange NaN values with infinity
        ResMS = (1./dof)*ResMS;
        ResMS(isnan(ResMS)) = Inf;
        
        % calclulate xBold
        xBold =  (contrast') * covM * contrast;
        
        % finally calculate variances
        VarB = xBold * ResMS;
        
       
        
        % get all the residuals data        
        %residualsOld = getResiduals(VRes,SPM.nscan,XYZ);
        for j=1:SPM.nscan,   
            residualsOld(j,:) = spm_get_data(VRes(j).fname,XYZ);
   
        end
        
        % calculate old variances of the residuals
        variancesOld = sum(residualsOld.^2,'omitnan');
        % 
        variancesOld = reshape(variancesOld, xDim, yDim, zDim);
        residualsOld = reshape(residualsOld, SPM.nscan, xDim, yDim, zDim);
        
        %============================
        % smooth Adaptive call
        %============================
        [betaN, ~, residualsNew] = smAd(gamma, VarB, residualsOld, hMax, lambda, voxelSizes',SPM.nscan);
        
     
        
        % transform gamma to 3D matrices
        %varNew = reshape(varN, xDim, yDim, zDim);
        betaNew = reshape(betaN, xDim, yDim, zDim);
        
        residualsNew = reshape(residualsNew,SPM.nscan,[]);
        
      
        
        % calculate new Variances of the residuals
        variancesNew = sum(residualsNew.^2,'omitnan');
        
         
        
         % standardize the residuals
        %Res1SS        = sum(residualsNew.^2,'omitnan');
        ResStandardized = zeros(SPM.nscan,xDim*yDim*zDim);
        for j=1:SPM.nscan,
            resJ=residualsNew(j,:);
            ss=sqrt(variancesNew/dof);
            ResStandardized(j,:) = resJ./ss; % /dof (xDim*yDim*zDim - 1)
  
        end
        
         ResStandardized=reshape(ResStandardized, SPM.nscan, xDim, yDim, zDim);
        % calculate reduction factor
        
         variancesNew = reshape(variancesNew, xDim, yDim, zDim);
        varRed = variancesNew./variancesOld ;
        
        % calculate new variances 
        varN = varRed.*VarB;
        
        % calculate ResMS from the variances computed with the new method 
        
        varNew = dof*varN;  
        varNew(isinf(varNew)) = NaN;
        ResMSNew = (1./xBold)*varNew;
        
%         varNew = dof*varNew;             %old, just for internal benchmark 
%         varNew(isinf(varNew)) = NaN;
%         ResMSNew = (1./xBold)*varNew;
        
        
        
         
        
        %-Save gamma and new ResMS
        
     
        
        %-Initialise beta image files (gamma)
        %----------------------------------------------------------------------
        VbetaNew = deal(struct('dim',      DIM', ...
            'dt',       [spm_type('float32') spm_platform('bigend')], ...
            'mat',      SPM.xY.VY(1).mat, ...
            'pinfo',    [1 0 0]'));
        VbetaNew.fname   = sprintf('gamma_smoothed_aws_%04d.nii', I);
        VbetaNew.descrip = sprintf('aws4SPM: gamma_%04d Smoothed SummedUp Beta Values!', I);
        
        
        % remove old file
        spm_unlink(VbetaNew.fname)
        VbetaNew = spm_create_vol(VbetaNew);
        
        
        % intialise residual sum of squares image file
        VResMSNew = struct('dim',		DIM',...
            'mat',		SPM.xY.VY(1).mat,...
            'pinfo',	[1 0 0]');
        VResMSNew.fname   = sprintf('ResMS_smoothed_aws_%04d.nii', I);
        VResMSNew.descrip = sprintf('aws4SPM: ResMS_smoothed_%04d Residual sum-of-squares', I);
        spm_unlink(VResMSNew.fname)
        VResMSNew = spm_create_vol(VResMSNew);
        
         % initialise standardised residual image file
        
        VResI(1:SPM.nscan) = deal(struct(...
                            'fname',   [],...
                            'dim',     DIM',...
                            'dt',      [spm_type('float64') spm_platform('bigend')],...
                            'mat',     SPM.xY.VY(1).mat,...
                            'pinfo',   [1 0 0]',...
                            'descrip', 'spm_spm:StandardisedResiduals'));
        
        % write gamma, ResMS image
        for z = 1:zDim
            VbetaNew = spm_write_plane(VbetaNew, betaNew(:,:,z), z);
            VResMSNew = spm_write_plane(VResMSNew, ResMSNew(:,:,z), z);
           % VResNew = spm_write_plane(VResNew, ResStandardized(:,:,z), z);
        end
        
        for i = 1:SPM.nscan
            VResI(i).fname   = [sprintf('ResI_aws_%04d', i) '.nii'];
            VResI(i).descrip = sprintf('spm_spm:ResI aws(%04d)', i);
        end
        VResI = spm_data_hdr_write(VResI);
        
        for j=1:SPM.nscan
              VResI(j) = spm_data_write(VResI(j), reshape(ResStandardized(j,:),DIM'));
        end
        
        % this is actually old SPM2, SPM99 code which should be no longer
        % needed!
        VbetaNew = spm_close_volSPM(VbetaNew);
        VResMSNew = spm_close_volSPM(VResMSNew);
        % VResNew = spm_close_volSPM(VResNew);
        
        % save volumes in the dedicated additional variables of SPM
        
        
        SPM.awsGamma(I) = VbetaNew;
        SPM.awsRes(I) = VResMSNew;
        
       
        % since spm_est_smooth overwrites the RPV.nii file, we copy it in
        % an RPV_original.nii file before and we will transfer it back
        % afterwards
        copyfile('RPV.nii','RPV_original.nii')
        
        
        
        % calculate the new FWHM, R and RPV File
         fprintf('Computing new FWHM and RPV... \n');
        [FWHM,VRpv,R] = spm_est_smoothness(VResI,SPM.VM,[SPM.nscan dof]); % SPM.awsRes(I)
        
        % copy files back in place and eliminate the copy
        awsRPV_fileName   = sprintf('RPV_smoothed_aws_%04d.nii', I);
        copyfile('RPV.nii',awsRPV_fileName) %put the correct number I
        copyfile('RPV_original.nii','RPV.nii')
        spm_unlink('RPV_original.nii')
        %spm_unlink('Res1st_smoothed_aws_0001.nii')
        
        %-Delete the standardised residual files
        %--------------------------------------------------------------------------
        for j=1:SPM.nscan,   
             spm_unlink(sprintf('ResI_aws_%04d.nii', j));   
        end
        
             
        
        VRpv.fname = awsRPV_fileName;
        
        % assign the values to the dedicated additional variables
        SPM.aws_FWHM{I} = FWHM;
        SPM.aws_VRpv(I) = VRpv;
        SPM.aws_R{I} = R;
        
        % save in the variables that are used by results
        SPM.xVol.FWHM =  FWHM;        
        SPM.xVol.R =  R;
        SPM.xVol.VRpv =  VRpv;
        
               
        
        [hreg,xSPM,SPM] = spm_results_aws('Setup',SPM,I);
        %         TabDat = spm_list('List',xSPM,hReg);
        
        
        
    end
    
    
    
    
    cd(swd);
    
    % assign the new additional variables to the original SPM
    originalSPM.aws_xCon = SPM.xCon;
    
    originalSPM.awsGamma = SPM.awsGamma;
    originalSPM.awsRes = SPM.awsRes;
    
    originalSPM.aws_FWHM = SPM.aws_FWHM;
    originalSPM.aws_VRpv = SPM.aws_VRpv;
    originalSPM.aws_R = SPM.aws_R;
    originalSPM.aws_VRes= VRes;
    
    % originalSPM.aws_ResM = ResM;
    
    % copy back the original SPM with the new variables in SPM 
    SPM = originalSPM;
    
    %-Save SPM.mat
    %--------------------------------------------------------------------------
    cd(swd);
    fprintf('%-40s: %30s','Saving SPM.mat','...writing');                   %-#
    save('SPM.mat','SPM', spm_get_defaults('mat.format'));
    fprintf('%s%30s\n',repmat(sprintf('\b'),1,30),'...done')                %-#
    cd(originalDir)
    
catch ME
    % if something goes wrong the original SPM is saved
    SPM=originalSPM;
    
    fprintf('%-40s: %30s','Saving SPM.mat','...writing');                   %-#
    save('SPM.mat','SPM', spm_get_defaults('mat.format'));
    fprintf('%s%30s\n',repmat(sprintf('\b'),1,30),'...done');
    hreg=[];
    xSPM=[];
    fprintf('Something went wrong during the execution of aws4SPM. No calculation done.\n')
    disp(getReport(ME));%,'basic'
end

function V = spm_close_volSPM(V)
% Close image volume
% See: spm_create_vol and spm_write_plane.
%_______________________________________________________________________
% @(#)spm_close_vol.m	2.4 John Ashburner 02/08/16

for i=1:prod(size(V)),
    if isfield(V,'private') & isfield(V(i).private,'fid') & ~isempty(V(i).private.fid),
        if ~isempty(fopen(V(i).private.fid)),
            fclose(V(i).private.fid);
        end;
        V(i).private = rmfield(V(i).private,'fid');
    end;
end;





